import { QuickQueue } from "@figliolia/data-structures";
/**
 * Priority Queue
 *
 * A bucket queue supporting a dynamic number of buckets
 * ```typescript
 * const queue = new PriorityQueue(3);
 * // queue = [[],[],[]]
 * queue.enqueue(() => {}, 1);
 * // queue = [[() => {}],[],[]]
 * queue.enqueue(() => {}, 2);
 * // queue = [[() => {}],[],[() => {}]]
 * queue.dequeue();
 * // queue = [[],[],[() => {}]]
 * queue.dequeue();
 * // queue = [[],[],[]]
 * ```
 */
export declare class PriorityQueue<T> {
    max: number;
    buckets: QuickQueue<T>[];
    constructor(buckets: number);
    /**
     * Enqueue
     *
     * Adds an item to the queue at the specified priority.
     * If no priority is provided, the item will be indexed
     * into the *highest* priority bucket
     */
    enqueue(item: T, priority?: number): string;
    /**
     * Dequeue
     *
     * Removes the first item in the priority queue and
     * returns it. Items are dequeued from higher priorities
     * before lower priorities
     */
    dequeue(): T | undefined;
    /**
     * Get Bucket
     *
     * Returns the Nth bucket in the Queue
     */
    getBucket(priority?: number): QuickQueue<T>;
    /**
     * peek
     *
     * Returns the highest priority item in the Queue
     */
    peek(): [string, T] | undefined;
    /**
     * IsEmpty
     *
     * Returns true if the Queue has no items in any of
     * its buckets
     */
    get isEmpty(): boolean;
    /**
     * Length
     *
     * Returns the total number of items in the Queue
     */
    get length(): number;
    /**
     * Delete By ID
     *
     * Removes an item from the Queue using its ID and
     * priority
     */
    delete(ID: string, priority: number): boolean;
    /**
     * Clear
     *
     * Resets the priority queue and removes all items
     */
    clear(): void;
    /**
     * Guard
     *
     * Throws a range error if the priority specified is out of range
     */
    private guard;
    [Symbol.iterator](): Generator<QuickQueue<T>, void, unknown>;
}
